from . import BaseAgent


class CounterModel(BaseAgent):
    """
    Dummy behaviour. It counts the number of nodes in the network and neighbors
    in each step and adds it to its state.
    """

    def step(self):
        # Outside effects
        total = len(list(self.get_all_agents()))
        neighbors = len(list(self.get_neighboring_agents()))
        self.state['times'] = self.state.get('times', 0) + 1
        self.state['neighbors'] = neighbors
        self.state['total'] = total


class AggregatedCounter(BaseAgent):
    """
    Dummy behaviour. It counts the number of nodes in the network and neighbors
    in each step and adds it to its state.
    """

    def step(self):
        # Outside effects
        total = len(list(self.get_all_agents()))
        neighbors = len(list(self.get_neighboring_agents()))
        self.state['times'] = self.state.get('times', 0) + 1
        self.state['neighbors'] = self.state.get('neighbors', 0) + neighbors
        self.state['total'] = total = self.state.get('total', 0) + total
        self.debug('Running for step: {}. Total: {}'.format(self.now, total))



import os
import sqlite3
import time
import weakref
import csv
import random
import simpy
from copy import deepcopy

import networkx as nx
import nxsim

from . import utils


class SoilEnvironment(nxsim.NetworkEnvironment):

    def __init__(self, name=None,
                 network_agents=None,
                 environment_agents=None,
                 states=None,
                 default_state=None,
                 interval=1,
                 seed=None,
                 dump=False,
                 simulation=None,
                 *args, **kwargs):
        self.name = name or 'UnnamedEnvironment'
        if isinstance(states, list):
            states = dict(enumerate(states))
        self.states = deepcopy(states) if states else {}
        self.default_state = deepcopy(default_state) or {}
        self.sim = weakref.ref(simulation)
        if 'topology' not in kwargs and simulation:
            kwargs['topology'] = self.sim().topology.copy()
        super().__init__(*args, **kwargs)
        self._env_agents = {}
        self.interval = interval
        self.dump = dump
        # Add environment agents first, so their events get
        # executed before network agents
        self['SEED'] = seed or time.time()
        random.seed(self['SEED'])
        self.process(self.save_state())
        self.environment_agents = environment_agents or []
        self.network_agents = network_agents or []
        if self.dump:
            self._db_path = os.path.join(self.get_path(), '{}.db.sqlite'.format(self.name))
        else:
            self._db_path = ":memory:"
        self.create_db(self._db_path)

    def create_db(self, db_path=None):
        db_path = db_path or self._db_path
        if os.path.exists(db_path):
            newname = db_path.replace('db.sqlite', 'backup{}.sqlite'.format(time.time()))
            os.rename(db_path, newname)
        self._db = sqlite3.connect(db_path)
        with self._db:
            self._db.execute('''CREATE TABLE IF NOT EXISTS history (agent_id text, t_step int, key text, value text, value_type text)''')

    @property
    def agents(self):
        yield from self.environment_agents
        yield from self.network_agents

    @property
    def environment_agents(self):
        for ref in self._env_agents.values():
            yield ref

    @environment_agents.setter
    def environment_agents(self, environment_agents):
        # Set up environmental agent
        self._env_agents = {}
        for item in environment_agents:
            kwargs = deepcopy(item)
            atype = kwargs.pop('agent_type')
            kwargs['agent_id'] = kwargs.get('agent_id', atype.__name__)
            kwargs['state'] = kwargs.get('state', {})
            a = atype(environment=self, **kwargs)
            self._env_agents[a.id] = a

    @property
    def network_agents(self):
        for i in self.G.nodes():
            node = self.G.node[i]
            if 'agent' in node:
                yield node['agent']

    @network_agents.setter
    def network_agents(self, network_agents):
        for ix in self.G.nodes():
            i = ix
            node = self.G.node[i]
            agent, state = utils.agent_from_distribution(network_agents)
            self.set_agent(i, agent_type=agent, state=state)

    def set_agent(self, agent_id, agent_type, state=None):
        node = self.G.nodes[agent_id]
        defstate = deepcopy(self.default_state)
        defstate.update(self.states.get(agent_id, {}))
        if state:
            defstate.update(state)
        state = defstate
        state.update(node.get('state', {}))
        a = agent_type(environment=self,
                       agent_id=agent_id,
                       state=state)
        node['agent'] = a
        return a

    def add_node(self, agent_type, state=None):
        agent_id = int(len(self.G.nodes()))
        self.G.add_node(agent_id)
        a = self.set_agent(agent_id, agent_type, state)
        a['visible'] = True
        return a

    def add_edge(self, agent1, agent2, attrs=None):
        return self.G.add_edge(agent1, agent2)

    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

    def _save_state(self, now=None):
        # for agent in self.agents:
        #     agent.save_state()
        utils.logger.debug('Saving state @{}'.format(self.now))
        with self._db:
            self._db.executemany("insert into history(agent_id, t_step, key, value, value_type) values (?, ?, ?, ?, ?)", self.state_to_tuples(now=now))

    def save_state(self):
        self._save_state()
        while self.peek() != simpy.core.Infinity:
            delay = max(self.peek() - self.now, self.interval)
            utils.logger.debug('Step: {}'.format(self.now))
            ev = self.event()
            ev._ok = True
            # Schedule the event with minimum priority so
            # that it executes before all agents
            self.schedule(ev, -999, delay)
            yield ev
            self._save_state()

    def __getitem__(self, key):
        if isinstance(key, tuple):
            values = [("agent_id", key[0]),
                      ("t_step", key[1]),
                      ("key", key[2]),
                      ("value", None),
                      ("value_type", None)]
            fields = list(k for k, v in values if v is None)
            conditions = " and ".join("{}='{}'".format(k, v) for k, v in values if v is not None)

            query = """SELECT {fields} from history""".format(fields=",".join(fields))
            if conditions:
                query = """{query} where {conditions}""".format(query=query,
                                                                conditions=conditions)
            with self._db:
                rows = self._db.execute(query).fetchall()

            utils.logger.debug(rows)
            results = self.rows_to_dict(rows)
            return results

        return self.environment_params[key]

    def rows_to_dict(self, rows):
        if len(rows) < 1:
            return None

        level = len(rows[0])-2

        if level == 0:
            if len(rows) != 1:
                raise ValueError('Cannot convert {} to dictionaries'.format(rows))
            value, value_type = rows[0]
            return utils.convert(value, value_type)

        results = {}
        for row in rows:
            item = results
            for i in range(level-1):
                key = row[i]
                if key not in item:
                    item[key] = {}
                item = item[key]
            key, value, value_type = row[level-1:]
            item[key] = utils.convert(value, value_type)
        return results

    def __setitem__(self, key, value):
        self.environment_params[key] = value

    def __contains__(self, key):
        return key in self.environment_params

    def get(self, key, default=None):
        return self[key] if key in self else default

    def get_path(self, dir_path=None):
        dir_path = dir_path or self.sim().dir_path
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        return dir_path

    def get_agent(self, agent_id):
        return self.G.node[agent_id]['agent']

    def get_agents(self):
        return list(self.agents)

    def dump_csv(self, dir_path=None):
        csv_name = os.path.join(self.get_path(dir_path),
                                '{}.environment.csv'.format(self.name))

        with open(csv_name, 'w') as f:
            cr = csv.writer(f)
            cr.writerow(('agent_id', 't_step', 'key', 'value', 'value_type'))
            for i in self.history_to_tuples():
                cr.writerow(i)

    def dump_gexf(self, dir_path=None):
        G = self.history_to_graph()
        graph_path = os.path.join(self.get_path(dir_path),
                                  self.name+".gexf")
        nx.write_gexf(G, graph_path, version="1.2draft")

    def state_to_tuples(self, now=None):
        if now is None:
            now = self.now
        for k, v in self.environment_params.items():
            v, v_t = utils.repr(v)
            yield 'env', now, k, v, v_t
        for agent in self.agents:
            for k, v in agent.state.items():
                v, v_t = utils.repr(v)
                yield agent.id, now, k, v, v_t

    def history_to_tuples(self):
        with self._db:
            res = self._db.execute("select agent_id, t_step, key, value, value_type from history ").fetchall()
        yield from res

    def history_to_graph(self):
        G = nx.Graph(self.G)

        for agent in self.network_agents:

            attributes = {'agent': str(agent.__class__)}
            lastattributes = {}
            spells = []
            lastvisible = False
            laststep = None
            history = self[agent.id, None, None]
            if not history:
                continue
            for t_step, state in reversed(sorted(list(history.items()))):
                for attribute, value in state.items():
                    if attribute == 'visible':
                        nowvisible = state[attribute]
                        if nowvisible and not lastvisible:
                            laststep = t_step
                        if not nowvisible and lastvisible:
                            spells.append((laststep, t_step))

                        lastvisible = nowvisible
                    else:
                        key = 'attr_' + attribute
                        if key not in attributes:
                            attributes[key] = list()
                        if key not in lastattributes:
                            lastattributes[key] = (state[attribute], t_step)
                        elif lastattributes[key][0] != value:
                            last_value, laststep = lastattributes[key]
                            value = (last_value, t_step, laststep)
                            if key not in attributes:
                                attributes[key] = list()
                            attributes[key].append(value)
                            lastattributes[key] = (state[attribute], t_step)
            for k, v in lastattributes.items():
                attributes[k].append((v[0], 0, v[1]))
            if lastvisible:
                spells.append((laststep, None))
            if spells:
                G.add_node(agent.id, spells=spells, **attributes)
            else:
                G.add_node(agent.id, **attributes)

        return G



import pandas as pd

import glob
import yaml
from os.path import join

from . import utils


def read_data(*args, group=False, **kwargs):
    iterable = _read_data(*args, **kwargs)
    if group:
        return group_trials(iterable)
    else:
        return list(iterable)


def _read_data(pattern, keys=None, convert_types=False,
               process=None, from_csv=False, **kwargs):
    for folder in glob.glob(pattern):
        config_file = glob.glob(join(folder, '*.yml'))[0]
        config = yaml.load(open(config_file))
        df = None
        if from_csv:
            for trial_data in sorted(glob.glob(join(folder,
                                                    '*.environment.csv'))):
                df = read_csv(trial_data, convert_types=convert_types)
                if process:
                    df = process(df, **kwargs)
                yield config_file, df, config
        else:
            for trial_data in sorted(glob.glob(join(folder, '*.db.sqlite'))):
                df = read_sql(trial_data, convert_types=convert_types,
                              keys=keys)
                if process:
                    df = process(df, **kwargs)
                yield config_file, df, config


def read_csv(filename, keys=None, convert_types=False, **kwargs):
    '''
    Read a CSV in canonical form: ::

        <agent_id, t_step, key, value, value_type>

    '''
    df = pd.read_csv(filename)
    if convert_types:
        df = convert_types_slow(df)
    if keys:
        df = df[df['key'].isin(keys)]
    return df


def read_sql(filename, keys=None, convert_types=False, limit=-1):
    condition = ''
    if keys:
        k = map(lambda x: "\'{}\'".format(x), keys)
        condition = 'where key in ({})'.format(','.join(k))
    query = 'select * from history {} limit {}'.format(condition, limit)
    df = pd.read_sql_query(query, 'sqlite:///{}'.format(filename))
    if convert_types:
        df = convert_types_slow(df)
    return df


def convert_row(row):
    row['value'] = utils.convert(row['value'], row['value_type'])
    return row


def convert_types_slow(df):
    '''This is a slow operation.'''
    dtypes = get_types(df)
    for k, v in dtypes.items():
        t = df[df['key']==k]
        t['value'] = t['value'].astype(v)
    df = df.apply(convert_row, axis=1)
    return df

def split_df(df):
    '''
    Split a dataframe in two dataframes: one with the history of agents,
    and one with the environment history
    '''
    envmask = (df['agent_id'] == 'env')
    n_env = envmask.sum()
    if n_env == len(df):
        return df, None
    elif n_env == 0:
        return None, df
    agents, env = [x for _, x in df.groupby(envmask)]
    return env, agents


def process(df, **kwargs):
    '''
    Process a dataframe in canonical form ``(t_step, agent_id, key, value, value_type)`` into
    two dataframes with a column per key: one with the history of the agents, and one for the
    history of the environment.
    '''
    env, agents = split_df(df)
    return process_one(env, **kwargs), process_one(agents, **kwargs)


def get_types(df):
    dtypes = df.groupby(by=['key'])['value_type'].unique()
    return {k:v[0] for k,v in dtypes.iteritems()}


def process_one(df, *keys, columns=['key'], values='value',
                index=['t_step', 'agent_id'], aggfunc='first', **kwargs):
    '''
    Process a dataframe in canonical form ``(t_step, agent_id, key, value, value_type)`` into
    a dataframe with a column per key
    '''
    if df is None:
        return df
    if keys:
        df = df[df['key'].isin(keys)]

    dtypes = get_types(df)

    df = df.pivot_table(values=values, index=index, columns=columns,
                        aggfunc=aggfunc, **kwargs)
    df = df.fillna(0).astype(dtypes)
    return df


def get_count_processed(df, *keys):
    if keys:
        df = df[list(keys)]
    # p = df.groupby(level=0).apply(pd.Series.value_counts)
    p = df.unstack().apply(pd.Series.value_counts, axis=1)
    return p


def get_count(df, *keys):
    if keys:
        df = df[df['key'].isin(keys)]
    p = df.groupby(by=['t_step', 'key', 'value']).size().unstack(level=[1,2]).fillna(0)
    return p


def get_value(df, *keys, aggfunc='sum'):
    if keys:
        df = df[df['key'].isin(keys)]
    p = process_one(df, *keys)
    p = p.groupby(level='t_step').agg(aggfunc)
    return p


def plot_all(*args, **kwargs):
    '''
    Read all the trial data and plot the result of applying a function on them.
    '''
    dfs = do_all(*args, **kwargs)
    ps = []
    for line in dfs:
        f, df, config = line
        df.plot(title=config['name'])
        ps.append(df)
    return ps

def do_all(pattern, func, *keys, include_env=False, **kwargs):
    for config_file, df, config in read_data(pattern, keys=keys):
        p = func(df, *keys, **kwargs)
        p.plot(title=config['name'])
        yield config_file, p, config


def group_trials(trials, aggfunc=['mean', 'min', 'max', 'std']):
    trials = list(trials)
    trials = list(map(lambda x: x[1] if isinstance(x, tuple) else x, trials))
    return pd.concat(trials).groupby(level=0).agg(aggfunc).reorder_levels([2, 0,1] ,axis=1)






import importlib
import sys
import os
import pdb
import logging

__version__ = "0.10.1"

try:
    basestring
except NameError:
    basestring = str

logging.basicConfig()

from . import agents
from . import simulation
from . import environment
from . import utils
from . import analysis


def main():
    import argparse
    from . import simulation

    parser = argparse.ArgumentParser(description='Run a SOIL simulation')
    parser.add_argument('file', type=str,
                        nargs="?",
                        default='simulation.yml',
                        help='python module containing the simulation configuration.')
    parser.add_argument('--module', '-m', type=str,
                        help='file containing the code of any custom agents.')
    parser.add_argument('--dry-run', '--dry', action='store_true',
                        help='Do not store the results of the simulation.')
    parser.add_argument('--pdb', action='store_true',
                        help='Use a pdb console in case of exception.')
    parser.add_argument('--output', '-o', type=str,
                        help='folder to write results to. It defaults to the current directory.')

    args = parser.parse_args()

    if args.module:
        sys.path.append(os.getcwd())
        importlib.import_module(args.module)

    logging.info('Loading config file: {}'.format(args.file, args.output))

    try:
        simulation.run_from_config(args.file, dump=(not args.dry_run), results_dir=args.output)
    except Exception as ex:
        if args.pdb:
            pdb.post_mortem()
        else:
            raise


if __name__ == '__main__':
    main()



from . import BaseAgent

import os.path
import matplotlib
import matplotlib.pyplot as plt
import networkx as nx


class DrawingAgent(BaseAgent):
    """
    Agent that draws the state of the network.
    """

    def step(self):
        # Outside effects
        f = plt.figure()
        nx.draw(self.env.G, node_size=10, width=0.2, pos=nx.spring_layout(self.env.G, scale=100), ax=f.add_subplot(111))
        f.savefig(os.path.join(self.env.sim().dir_path, "graph-"+str(self.env.now)+".png"))



import os
import yaml
import logging
from time import time
from glob import glob
from random import random
from copy import deepcopy

import networkx as nx

from contextlib import contextmanager


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def load_network(network_params, dir_path=None):
    if network_params is None:
        return nx.Graph()
    path = network_params.get('path', None)
    if path:
        if dir_path and not os.path.isabs(path):
            path = os.path.join(dir_path, path)
        extension = os.path.splitext(path)[1][1:]
        kwargs = {}
        if extension == 'gexf':
            kwargs['version'] = '1.2draft'
            kwargs['node_type'] = int
        try:
            method = getattr(nx.readwrite, 'read_' + extension)
        except AttributeError:
            raise AttributeError('Unknown format')
        return method(path, **kwargs)

    net_args = network_params.copy()
    net_type = net_args.pop('generator')

    method = getattr(nx.generators, net_type)
    return method(**net_args)


def load_file(infile):
    with open(infile, 'r') as f:
        return list(yaml.load_all(f))


def load_files(*patterns):
    for pattern in patterns:
        for i in glob(pattern):
            for config in load_file(i):
                yield config, os.path.abspath(i)


def load_config(config):
    if isinstance(config, dict):
        yield config, None
    else:
        yield from load_files(config)


@contextmanager
def timer(name='task', pre="", function=logger.info, to_object=None):
    start = time()
    yield start
    end = time()
    function('{}Finished {} in {} seconds'.format(pre, name, str(end-start)))
    if to_object:
        to_object.start = start
        to_object.end = end


def agent_from_distribution(distribution, value=-1):
    """Find the agent """
    if value < 0:
        value = random()
    for d in distribution:
        threshold = d['threshold']
        if value >= threshold[0] and value < threshold[1]:
            state = {}
            if 'state' in d:
                state = deepcopy(d['state'])
            return d['agent_type'], state

    raise Exception('Distribution for value {} not found in: {}'.format(value, distribution))


def repr(v):
    if isinstance(v, bool):
        v = "true" if v else ""
        return v, bool.__name__
    return v, type(v).__name__

def convert(value, type_):
    import importlib
    try:
        # Check if it's a builtin type
        module = importlib.import_module('builtins')
        cls = getattr(module, type_)
    except AttributeError:
        # if not, separate module and class
        module, type_ = type_.rsplit(".", 1)
        module = importlib.import_module(module)
        cls = getattr(module, type_)
    return cls(value)



import os
import time
import imp
import sys
import yaml
import networkx as nx
from networkx.readwrite import json_graph

from copy import deepcopy

import pickle

from nxsim import NetworkSimulation

from . import agents, utils, environment, basestring
from .utils import logger


class SoilSimulation(NetworkSimulation):
    """
    Subclass of nsim.NetworkSimulation with three main differences:
        1) agent type can be specified by name or by class.
        2) instead of just one type, an network_agents can be used.
           The distribution specifies the weight (or probability) of each
           agent type in the topology. This is an example distribution: ::

                  [
                    {'agent_type': 'agent_type_1',
                     'weight': 0.2,
                     'state': {
                         'id': 0
                      }
                    },
                    {'agent_type': 'agent_type_2',
                     'weight': 0.8,
                     'state': {
                         'id': 1
                      }
                    }
                  ]

          In this example, 20% of the nodes will be marked as type
          'agent_type_1'.
        3) if no initial state is given, each node's state will be set
           to `{'id': 0}`.
    """
    def __init__(self, name=None, topology=None, network_params=None,
                 network_agents=None, agent_type=None, states=None,
                 default_state=None, interval=1, dump=False,
                 dir_path=None, num_trials=1, max_time=100,
                 agent_module=None, load_module=None, seed=None,
                 environment_agents=None, environment_params=None):

        if topology is None:
            topology = utils.load_network(network_params,
                                          dir_path=dir_path)
        elif isinstance(topology, basestring) or isinstance(topology, dict):
            topology = json_graph.node_link_graph(topology)


        self.load_module = load_module
        self.topology = nx.Graph(topology)
        self.network_params = network_params
        self.name = name or 'UnnamedSimulation'
        self.num_trials = num_trials
        self.max_time = max_time
        self.default_state = default_state or {}
        self.dir_path = dir_path or os.getcwd()
        self.interval = interval
        self.seed = str(seed) or str(time.time())
        self.dump = dump
        self.environment_params = environment_params or {}

        if load_module:
            path = sys.path + [self.dir_path]
            f, fp, desc = imp.find_module(load_module, path)
            imp.load_module('soil.agents.custom', f, fp, desc)

        environment_agents = environment_agents or []
        self.environment_agents = self._convert_agent_types(environment_agents)

        distro = self.calculate_distribution(network_agents,
                                             agent_type)
        self.network_agents = self._convert_agent_types(distro)

        self.states = self.validate_states(states,
                                           self.topology)

    def calculate_distribution(self,
                               network_agents=None,
                               agent_type=None):
        if network_agents:
            network_agents = deepcopy(network_agents)
        elif agent_type:
            network_agents = [{'agent_type': agent_type}]
        else:
            return []

        # Calculate the thresholds
        total = sum(x.get('weight', 1) for x in network_agents)
        acc = 0
        for v in network_agents:
            upper = acc + (v.get('weight', 1)/total)
            v['threshold'] = [acc, upper]
            acc = upper
        return network_agents

    def serialize_distribution(self):
        d = self._convert_agent_types(self.network_agents,
                                      to_string=True)
        for v in d:
            if 'threshold' in v:
                del v['threshold']
        return d

    def _convert_agent_types(self, ind, to_string=False):
        d = deepcopy(ind)
        for v in d:
            agent_type = v['agent_type']
            if to_string and not isinstance(agent_type, str):
                v['agent_type'] = str(agent_type.__name__)
            elif not to_string and isinstance(agent_type, str):
                v['agent_type'] = agents.agent_types[agent_type]
        return d

    def validate_states(self, states, topology):
        states = states or []
        # Validate states to avoid ignoring states during
        # initialization
        if isinstance(states, dict):
            for x in states:
                assert x in self.topology.node
        else:
            assert len(states) <= len(self.topology)
        return states

    def run_simulation(self):
        return self.run()

    def run(self):
        return list(self.run_simulation_gen())

    def run_simulation_gen(self, *args, **kwargs):
        with utils.timer('simulation'):
            for i in range(self.num_trials):
                res = self.run_trial(i)
                if self.dump:
                    res.dump_gexf(self.dir_path)
                    res.dump_csv(self.dir_path)
                yield res

            if self.dump:
                logger.info('Dumping results to {}'.format(self.dir_path))
                self.dump_pickle(self.dir_path)
                self.dump_yaml(self.dir_path)
            else:
                logger.info('NOT dumping results')

    def run_trial(self, trial_id=0, dump=False, dir_path=None):
        """Run a single trial of the simulation

        Parameters
        ----------
        trial_id : int
        """
        # Set-up trial environment and graph
        logger.info('Trial: {}'.format(trial_id))
        env_name = '{}_trial_{}'.format(self.name, trial_id)
        env = environment.SoilEnvironment(name=env_name,
                                          topology=self.topology.copy(),
                                          seed=self.seed+env_name,
                                          initial_time=0,
                                          dump=self.dump,
                                          interval=self.interval,
                                          network_agents=self.network_agents,
                                          states=self.states,
                                          default_state=self.default_state,
                                          environment_agents=self.environment_agents,
                                          simulation=self,
                                          **self.environment_params)
        # Set up agents on nodes
        logger.info('\tRunning')
        with utils.timer('trial'):
            env.run(until=self.max_time)
        return env

    def to_dict(self):
        return self.__getstate__()

    def to_yaml(self):
        return yaml.dump(self.to_dict())

    def dump_yaml(self, dir_path=None, file_name=None):
        dir_path = dir_path or self.dir_path
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        if not file_name:
            file_name = os.path.join(dir_path,
                                     '{}.dumped.yml'.format(self.name))
        with open(file_name, 'w') as f:
            f.write(self.to_yaml())

    def dump_pickle(self, dir_path=None, pickle_name=None):
        dir_path = dir_path or self.dir_path
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        if not pickle_name:
            pickle_name = os.path.join(dir_path,
                                       '{}.simulation.pickle'.format(self.name))
        with open(pickle_name, 'wb') as f:
            pickle.dump(self, f)

    def __getstate__(self):
        state = self.__dict__.copy()
        state['topology'] = json_graph.node_link_data(self.topology)
        state['network_agents'] = self.serialize_distribution()
        state['environment_agents'] = self._convert_agent_types(self.environment_agents,
                                                          to_string=True)
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self.topology = json_graph.node_link_graph(state['topology'])
        self.network_agents = self._convert_agent_types(self.network_agents)
        self.environment_agents = self._convert_agent_types(self.environment_agents)
        return state


def from_config(config, G=None):
    config = list(utils.load_config(config))
    if len(config) > 1:
        raise AttributeError('Provide only one configuration')
    config = config[0][0]
    sim = SoilSimulation(**config)
    return sim


def run_from_config(*configs, dump=True, results_dir=None, timestamp=False):
    if not results_dir:
        results_dir = 'soil_output'
    for config_def in configs:
        for config, cpath in utils.load_config(config_def):
            name = config.get('name', 'unnamed')
            logger.info("Using config(s): {name}".format(name=name))

            sim = SoilSimulation(**config)
            if timestamp:
                sim_folder = '{}_{}'.format(sim.name,
                                            time.strftime("%Y-%m-%d_%H:%M:%S"))
            else:
                sim_folder = sim.name
            sim.dir_path = os.path.join(results_dir, sim_folder)
            sim.dump = dump
            logger.info('Dumping results to {} : {}'.format(sim.dir_path, dump))
            results = sim.run_simulation()



# networkStatus = {}  # Dict that will contain the status of every agent in the network
# sentimentCorrelationNodeArray = []
# for x in range(0, settings.network_params["number_of_nodes"]):
#     sentimentCorrelationNodeArray.append({'id': x})
# Initialize agent states. Let's assume everyone is normal.
    

import nxsim
import logging
from collections import OrderedDict
from copy import deepcopy
from functools import partial
import json


from functools import wraps


agent_types = {}


class MetaAgent(type):
    def __init__(cls, name, bases, nmspc):
        super(MetaAgent, cls).__init__(name, bases, nmspc)
        agent_types[name] = cls


class BaseAgent(nxsim.BaseAgent, metaclass=MetaAgent):
    """
    A special simpy BaseAgent that keeps track of its state history.
    """

    defaults = {}

    def __init__(self, **kwargs):
        self._neighbors = None
        self.alive = True
        state = deepcopy(self.defaults)
        state.update(kwargs.pop('state', {}))
        kwargs['state'] = state
        super().__init__(**kwargs)
        if not hasattr(self, 'level'):
            self.level = logging.DEBUG
        self.logger = logging.getLogger('Agent-{}'.format(self.id))
        self.logger.setLevel(self.level)


    def __getitem__(self, key):
        if isinstance(key, tuple):
            k, t_step = key
            return self.env[self.id, t_step, k]
        return self.state.get(key, None)

    def __delitem__(self, key):
        del self.state[key]

    def __contains__(self, key):
        return key in self.state

    def __setitem__(self, key, value):
        self.state[key] = value

    def get(self, key, default=None):
        return self[key] if key in self else default

    @property
    def now(self):
        try:
            return self.env.now
        except AttributeError:
            # No environment
            return None

    def run(self):
        interval = self.env.interval
        while self.alive:
            res = self.step()
            yield res or self.env.timeout(interval)

    def die(self, remove=False):
        self.alive = False
        if remove:
            super().die()

    def step(self):
        pass

    def to_json(self):
        return json.dumps(self.state)

    def count_agents(self, state_id=None, limit_neighbors=False):
        if limit_neighbors:
            agents = self.global_topology.neighbors(self.id)
        else:
            agents = self.global_topology.nodes()
        count = 0
        for agent in agents:
            if state_id and state_id != self.global_topology.node[agent]['agent'].state['id']:
                continue
            count += 1
        return count

    def count_neighboring_agents(self, state_id=None):
        return len(super().get_agents(state_id, limit_neighbors=True))

    def get_agents(self, state_id=None, limit_neighbors=False, iterator=False, **kwargs):
        if limit_neighbors:
            agents = super().get_agents(state_id, limit_neighbors)
        else:
            agents = filter(lambda x: state_id is None or x.state.get('id', None) == state_id,
                            self.env.agents)

        def matches_all(agent):
            state = agent.state
            for k, v in kwargs.items():
                if state.get(k, None) != v:
                    return False
            return True

        f = filter(matches_all, agents)
        if iterator:
            return f
        return list(f)

    def log(self, message, *args, level=logging.INFO, **kwargs):
        message = message + " ".join(str(i) for i in args)
        message = "\t@{:>5}:\t{}".format(self.now, message)
        for k, v in kwargs:
            message += " {k}={v} ".format(k, v)
        extra = {}
        extra['now'] = self.now
        extra['id'] = self.id
        return self.logger.log(level, message, extra=extra)

    def debug(self, *args, **kwargs):
        return self.log(*args, level=logging.DEBUG, **kwargs)

    def info(self, *args, **kwargs):
        return self.log(*args, level=logging.INFO, **kwargs)


def state(func):

    @wraps(func)
    def func_wrapper(self):
        when = None
        next_state = func(self)
        try:
            next_state, when = next_state
        except TypeError:
            pass
        if next_state:
            try:
                self.state['id'] = next_state.id
            except AttributeError:
                raise ValueError('State id %s is not valid.' % next_state)
        return when

    func_wrapper.id = func.__name__
    func_wrapper.is_default = False
    return func_wrapper


def default_state(func):
    func.is_default = True
    return func


class MetaFSM(MetaAgent):
    def __init__(cls, name, bases, nmspc):
        super(MetaFSM, cls).__init__(name, bases, nmspc)
        states = {}
        # Re-use states from inherited classes
        default_state = None
        for i in bases:
            if isinstance(i, MetaFSM):
                for state_id, state in i.states.items():
                    if state.is_default:
                        default_state = state
                    states[state_id] = state

        # Add new states
        for name, func in nmspc.items():
            if hasattr(func, 'id'):
                if func.is_default:
                    default_state = func
                states[func.id] = func
        cls.default_state = default_state
        cls.states = states


class FSM(BaseAgent, metaclass=MetaFSM):
    def __init__(self, *args, **kwargs):
        super(FSM, self).__init__(*args, **kwargs)
        if 'id' not in self.state:
            self.state['id'] = self.default_state.id

    def step(self):
        if 'id' in self.state:
            next_state = self.state['id']
        elif self.default_state:
            next_state = self.default_state.id
        else:
            raise Exception('{} has no valid state id or default state'.format(self))
        if next_state not in self.states:
            raise Exception('{} is not a valid id for {}'.format(next_state, self))
        self.states[next_state](self)

    def set_state(self, state):
        if hasattr(state, 'id'):
            state = state.id
        if state not in self.states:
            raise ValueError('{} is not a valid state'.format(state))
        self.state['id'] = state


from .BassModel import *
from .BigMarketModel import *
from .IndependentCascadeModel import *
from .ModelM2 import *
from .SentimentCorrelationModel import *
from .SISaModel import *
from .CounterModel import *
from .DrawingAgent import *



import pip
from setuptools import setup
# parse_requirements() returns generator of pip.req.InstallRequirement objects
from pip.req import parse_requirements
from soil import __version__

try:
    install_reqs = parse_requirements(
        "requirements.txt", session=pip.download.PipSession())
    test_reqs = parse_requirements(
        "test-requirements.txt", session=pip.download.PipSession())
except AttributeError:
    install_reqs = parse_requirements("requirements.txt")
    test_reqs = parse_requirements("test-requirements.txt")

install_reqs = [str(ir.req) for ir in install_reqs]
test_reqs = [str(ir.req) for ir in test_reqs]


setup(
    name='soil',
    packages=['soil'],  # this must be the same as the name above
    version=__version__,
    description=('An Agent-Based Social Simulator for Social Networks'),
    author='J. Fernando Sanchez',
    author_email='jf.sanchez@upm.es',
    url='https://github.com/gsi-upm/soil',  # use the URL to the github repo
    download_url='https://github.com/gsi-upm/soil/archive/{}.tar.gz'.format(
        __version__),
    keywords=['agent', 'social', 'simulator'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3'],
    install_requires=install_reqs,
    tests_require=test_reqs,
    setup_requires=['pytest-runner', ],
    include_package_data=True,
    entry_points={
        'console_scripts':
        ['soil = soil.__init__:main']
    })
